package de.howaner.Pokemon;

import com.google.gson.Gson;
import de.howaner.Pokemon.command.CommandManager;
import de.howaner.Pokemon.config.DatabaseConfig;
import de.howaner.Pokemon.console.TerminalConsoleWriterThread;
import de.howaner.Pokemon.config.ServerConfig;
import de.howaner.Pokemon.console.ThreadCommandReader;
import de.howaner.Pokemon.database.Database;
import de.howaner.Pokemon.database.DatabaseCallback;
import de.howaner.Pokemon.database.DatabaseManager;
import de.howaner.Pokemon.database.DatabasePlayer;
import de.howaner.Pokemon.entity.Player;
import de.howaner.Pokemon.listener.EventManager;
import de.howaner.Pokemon.listener.event.player.PlayerQuitEvent;
import de.howaner.Pokemon.network.netty.MessageDeserializer;
import de.howaner.Pokemon.network.netty.MessageLengthDeserializer;
import de.howaner.Pokemon.network.netty.MessageLengthSerializer;
import de.howaner.Pokemon.network.netty.MessageSerializer;
import de.howaner.Pokemon.network.NetworkManager;
import de.howaner.Pokemon.network.packets.out.OutPacket;
import de.howaner.Pokemon.network.util.CryptManager;
import de.howaner.Pokemon.world.World;
import de.howaner.Pokemon.world.WorldLoader;
import de.howaner.Pokemon.world.WorldManager;
import io.netty.bootstrap.ServerBootstrap;
import io.netty.channel.Channel;
import io.netty.channel.ChannelException;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.ChannelOption;
import io.netty.channel.socket.nio.NioServerSocketChannel;
import io.netty.handler.timeout.ReadTimeoutHandler;
import java.io.File;
import java.net.InetAddress;
import java.security.KeyPair;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import jline.console.ConsoleReader;
import lombok.Getter;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.apache.logging.log4j.core.Appender;
import org.apache.logging.log4j.core.appender.ConsoleAppender;

public class PokemonServer implements Runnable {
	private static PokemonServer instance;
	private boolean serverRunning = true;
	private volatile boolean isInStopping = false;
	private Thread serverThread;
	private ConsoleReader consoleReader;
	private CommandManager commandManager;
	private KeyPair serverKey;
	@Getter private DatabaseManager databaseManager;
	private Map<String, Player> onlinePlayers = new HashMap<String, Player>();

	@Getter private Gson gson = new Gson();
	private Logger logger;
	private Scheduler scheduler;
	private ServerConfig configuration;
	private WorldManager worldManager;
	private EventManager eventManager;

	// --- Network Stuff --- //
	private final List<NetworkManager> netManagers = Collections.synchronizedList(new ArrayList<NetworkManager>());
	private ChannelFuture endpoint;

	// --- "GET" Methods --- //

	public static PokemonServer getServer() {
		return instance;
	}

	public Logger getLogger() {
		return this.logger;
	}

	public boolean isServerThread() {
		return (Thread.currentThread() == this.serverThread);
	}

	public Thread getServerThread() {
		return this.serverThread;
	}

	public Scheduler getScheduler() {
		return this.scheduler;
	}

	public ServerConfig getConfig() {
		return this.configuration;
	}

	public ConsoleReader getConsoleReader() {
		return this.consoleReader;
	}

	public CommandManager getCommandManager() {
		return this.commandManager;
	}

	public WorldManager getWorldManager() {
		return this.worldManager;
	}

	public EventManager getEventManager() {
		return this.eventManager;
	}

	public KeyPair getServerKey() {
		return this.serverKey;
	}

	public boolean isRunning() {
		return this.serverRunning;
	}

	public Collection<Player> getOnlinePlayers() {
		return this.onlinePlayers.values();
	}

	public Player getPlayer(String name) {
		return this.onlinePlayers.get(name);
	}

	// --- Class Stuff --- //

	public PokemonServer() {
		// Set instance
		instance = this;

		// Initialize logger
		{
			org.apache.logging.log4j.core.Logger rootLogger = (org.apache.logging.log4j.core.Logger) LogManager.getRootLogger();
			for (Appender appender : rootLogger.getAppenders().values()) {
				if (appender instanceof ConsoleAppender) {
					rootLogger.removeAppender(appender);
					break;
				}
			}
		}
		this.logger = LogManager.getLogger();

		this.logger.info("------ Start server at {} ------", new SimpleDateFormat().format(Calendar.getInstance().getTime()));
		this.logger.info("Starting server ...");
		this.startServerThread();
	}

	private void initialize() {
		// Load config
		try {
			this.configuration = new ServerConfig();
			this.configuration.readConfig();
		} catch (Exception ex) {
			logger.error("Can't load config file", ex);
		}

		// Load command manager
		this.commandManager = new CommandManager();

		// Load jline
		boolean isJLineSupported = !"jline.UnsupportedTerminal".equals(System.getProperty("jline.terminal"));
		if (!isJLineSupported || !this.getConfig().isJLineEnabled() || System.console() == null) {
			System.setProperty("jline.terminal", "jline.UnsupportedTerminal");
			System.setProperty("user.language", "en");
			this.getConfig().setUseJLine(false);
		}

		// Load console reader
		try {
			this.consoleReader = new ConsoleReader(System.in, System.out);
			this.consoleReader.setExpandEvents(false);
		} catch (Exception ex) {
			try {
				System.setProperty("jline.terminal", "jline.UnsupportedTerminal");
				System.setProperty("user.language", "en");
				this.getConfig().setUseJLine(false);

				this.consoleReader = new ConsoleReader(System.in, System.out);
				this.consoleReader.setExpandEvents(false);
			} catch (Exception exx) {
				this.logger.error("Failed to initialize console reader", exx);
			}
		}

		// Load command reader thread
		{
			ThreadCommandReader commandReaderThread = new ThreadCommandReader();
			commandReaderThread.setDaemon(true);
			commandReaderThread.start();
		}

		new Thread(new TerminalConsoleWriterThread(System.out, this.consoleReader)).start();

		// Load keys
		try {
			this.loadServerKey();
		} catch (Exception ex) {
			this.logger.error("Can't load private and public key", ex);
			this.shutdown();
			return;
		}

		// Load world manager
		try {
			this.logger.info("Load main world ...");
			this.worldManager = new WorldManager();
			this.worldManager.loadMainWorld();
			this.logger.info("Loaded main world!");
		} catch (Exception ex) {
			this.logger.error("Can't load main world.", ex);
			this.shutdown();
			return;
		}

		// Connect to database
		Database db = new Database();
		try {
			this.logger.info("Connect to mongodb ...");
			long timestamp = System.currentTimeMillis();

			DatabaseConfig cfg = this.configuration.getDatabaseConfig();
			db.connect(cfg.getMongoIP(), cfg.getMongoPort(), cfg.getMongoUser(), cfg.getMongoPass(), cfg.getDatabase());

			this.logger.info("Connected to mongodb in {}ms!", System.currentTimeMillis() - timestamp);
		} catch (Exception ex) {
			this.logger.error("Can't connect to mongodb", ex);
			this.shutdown();
			return;
		}

		this.databaseManager = new DatabaseManager(db);
		this.eventManager = new EventManager();
		this.scheduler = new Scheduler();
		this.worldManager.registerStandardListeners();

		// Start tcp server
		try {
			this.startNetwork();
		} catch (Exception ex) {
			this.logger.error("Failed to bind to ip address!!!!", ex);
			this.shutdown();
			return;
		}
	}

	public void tick() {
		this.scheduler.tick();
		this.commandManager.executeQueuedCommands();
		this.networkTick();
		this.worldManager.onTick();
	}

	@Override
	public void run() {
		try {
			this.initialize();

			long lastTime = System.currentTimeMillis();
			long overloadedTime = 0L;

			while (this.serverRunning) {
				long currentTime = System.currentTimeMillis();
				long pastTime = currentTime - lastTime;

				if (pastTime > 2000L) {
					// Too high past time -> Set past time to 2000
					pastTime = 2000L;
				}

				if (pastTime < 0L) {
					// What the fuck? Backwards running?
					pastTime = 0L;
				}

				overloadedTime += pastTime;
				lastTime = currentTime;

				while (overloadedTime > 50L) {
					overloadedTime -= 50L;
					this.tick();
				}

				Thread.sleep(Math.max(1L, 50L - overloadedTime));
			}
		} catch (Exception ex) {
			this.logger.error("Encountered an unexpected exception", ex);
		} finally {
			this.stopServerThread();
			try {
				this.consoleReader.getTerminal().restore();
			} catch (Exception e) {
			}
		}
	}

	public void startServerThread() {
		if (this.serverThread != null && this.serverThread.isAlive()) {
			throw new RuntimeException("Server thread was already started");
		}

		this.serverThread = new Thread(this, "Server thread");
		this.serverThread.start();
	}

	public void stopServerThread() {
		this.serverRunning = false;

		if (this.isInStopping) {
			return;
		}
		this.isInStopping = true;

		this.logger.info("Stopping server");
	}

	public void shutdown() {
		this.stopServerThread();

		this.stopNetwork();

		try {
			this.consoleReader.getTerminal().restore();
		} catch (Exception e) { }

		System.exit(0);
	}

	private void loadServerKey() throws Exception {
		if (this.serverKey != null) {
			//Server key is already loaded
			return;
		}

		File keysFolder = new File("keys");
		if (!keysFolder.exists()) {
			keysFolder.mkdir();
		}

		File privateKeyFile = new File(keysFolder, "private.key");
		File publicKeyFile = new File(keysFolder, "public.key");

		if (privateKeyFile.exists() && publicKeyFile.exists()) {
			PrivateKey privateKey = CryptManager.readPrivateKeyFromFile(privateKeyFile);
			PublicKey publicKey = CryptManager.readPublicKeyFromFile(publicKeyFile);

			this.serverKey = new KeyPair(publicKey, privateKey);
			this.logger.info("Loaded private and public key!");
			return;
		}

		//Create keys
		this.serverKey = CryptManager.createKeyPair();
		CryptManager.savePrivateKey(this.serverKey.getPrivate(), privateKeyFile);
		CryptManager.savePublicKey(this.serverKey.getPublic(), publicKeyFile);
		this.logger.info("Created new private and public key!");
	}

	// --- Network Stuff --- //
	public void startNetwork() throws Exception {
		final InetAddress address = InetAddress.getByName(this.getConfig().getServerIP());
		final int port = this.getConfig().getServerPort();

		this.endpoint = new ServerBootstrap().channel(NioServerSocketChannel.class).childHandler(new ChannelInitializer() {
			@Override
			protected void initChannel(Channel channel) throws Exception {
				try {
					channel.config().setOption(ChannelOption.IP_TOS, 24);
				} catch (ChannelException ex) {}

				try {
					channel.config().setOption(ChannelOption.TCP_NODELAY, false);
				} catch (ChannelException ex) {}

				channel.pipeline().addLast("timeout", new ReadTimeoutHandler(30));
				channel.pipeline().addLast("splitter", new MessageLengthDeserializer());
				channel.pipeline().addLast("decoder", new MessageDeserializer());
				channel.pipeline().addLast("prepender", new MessageLengthSerializer());
				channel.pipeline().addLast("encoder", new MessageSerializer());

				NetworkManager netManager = new NetworkManager();
				channel.pipeline().addLast("packet_handler", netManager);
				synchronized (PokemonServer.this.netManagers) {
					PokemonServer.this.netManagers.add(netManager);
				}
				PokemonServer.getServer().getLogger().info("Incoming connection from {}", channel.remoteAddress());
			}
		}).group(NetworkManager.eventLoops).localAddress(address, port).bind().syncUninterruptibly();
		this.logger.info("Started tcp server at {}:{}", address.getHostName(), port);
	}

	public void stopNetwork() {
		if (this.endpoint == null) {
			return;
		}

		this.endpoint.channel().close().syncUninterruptibly();
	}

	public void networkTick() {
		synchronized (this.netManagers) {
			Iterator<NetworkManager> itr = this.netManagers.iterator();
			while (itr.hasNext()) {
				final NetworkManager netManager = itr.next();

				if (!netManager.isChannelOpen()) {
					itr.remove();

					if (netManager.getDisconnectMessage().isEmpty()) {
						netManager.getClientHandle().onDisconnect("Disconnected");
					} else {
						netManager.getClientHandle().onDisconnect(netManager.getDisconnectMessage());
					}
				} else {
					try {
						netManager.processQueue();
					} catch (Exception ex) {
						this.logger.warn("Failed to handle packets for " + netManager.getRemoteAddress(), ex);
						netManager.getClientHandle().kick("Big packet error!");
					}
				}
			}
		}
	}

	public void onPlayerJoin(final Player player) {
		if (this.getWorldManager().getWorld(player.getLocation().getWorld()) != null) {
			World world = this.getWorldManager().getWorld(player.getLocation().getWorld());
			world.spawnEntity(player);
			this.onlinePlayers.put(player.getName(), player);
			player.getClientHandle().onLoginSuccess();
		} else {
			// -> Load world
			this.getWorldManager().getWorldLoader().addWorldToQueue(player.getLocation().getWorld(), new WorldLoader.WorldLoadedCallback() {
				@Override
				public void onLoadedWorld(World world) {
					world.spawnEntity(player);
					PokemonServer.this.onlinePlayers.put(player.getName(), player);
					player.getClientHandle().onLoginSuccess();
				}
			});
		}
	}

	public void onPlayerLeave(final Player player) {
		this.getLogger().info("Player {} left the game.", player.getName());
		this.onlinePlayers.remove(player.getName());
		PokemonServer.getServer().getEventManager().callEvent(new PlayerQuitEvent(player));
		player.getWorld().removeEntity(player.getEntityID());

		// Save player data
		this.getDatabaseManager().getPlayer(player.getName(), new DatabaseCallback<DatabasePlayer>() {
			@Override
			public void run(DatabasePlayer dbPlayer) {
				if (dbPlayer == null) return;

				dbPlayer.setLocation(player.getLocation());
				PokemonServer.this.getDatabaseManager().setPlayer(dbPlayer);
			}
		});
	}

	public void broadcastPacket(OutPacket packet) {
		for (NetworkManager netManager : this.netManagers) {
			if (netManager.isChannelOpen())
				netManager.getClientHandle().sendPacket(packet);
		}
	}

}
