package de.howaner.Pokemon.network;

import com.google.common.collect.Queues;
import de.howaner.Pokemon.PokemonServer;
import de.howaner.Pokemon.database.DatabaseCallback;
import de.howaner.Pokemon.database.DatabasePlayer;
import de.howaner.Pokemon.entity.Player;
import de.howaner.Pokemon.listener.event.player.PlayerJoinEvent;
import de.howaner.Pokemon.network.packets.in.*;
import de.howaner.Pokemon.network.packets.out.*;
import de.howaner.Pokemon.network.util.CryptManager;
import de.howaner.Pokemon.util.Location;
import io.netty.util.concurrent.Future;
import io.netty.util.concurrent.GenericFutureListener;
import java.util.Queue;
import javax.crypto.Cipher;
import javax.crypto.SecretKey;
import javax.crypto.spec.SecretKeySpec;

public class ClientHandle {
	private final NetworkManager netManager;
	private long networkTickCount = 0L;
	private long lastKeepAliveTick = 0L;
	private ConnectionState state = ConnectionState.NOTHING;

	private Player player;
	private String userName = "";
	private Queue<InPacketMove> movePacketsQueue = Queues.newConcurrentLinkedQueue();

	public ClientHandle(NetworkManager netManager) {
		this.netManager = netManager;
	}

	public Player getPlayer() {
		return this.player;
	}

	public void setPlayer(Player player) {
		this.player = player;
	}

	public String getUserName() {
		return this.userName;
	}

	public NetworkManager getNetworkManager() {
		return this.netManager;
	}
	
	public void onNetworkTick() {
		this.networkTickCount++;

		// Keep Alive
		if ((this.networkTickCount - this.lastKeepAliveTick) > 40L) {
			this.lastKeepAliveTick = this.networkTickCount;
			this.sendPacket(new PacketOutKeepAlive(this.lastKeepAliveTick));
		}

		while (!this.movePacketsQueue.isEmpty() && !this.player.isInMove()) {
			InPacketMove packet = this.movePacketsQueue.poll();
			if (packet instanceof PacketInPlayerMove)
				this.handleMove((PacketInPlayerMove) packet);
			else if (packet instanceof PacketInChangeFace)
				this.handleChangeFace((PacketInChangeFace) packet);
			else if (packet instanceof PacketInChangeState)
				this.handleChangeState((PacketInChangeState) packet);
		}
	}

	// ----- MOVE METHODS ----- //
	public void clearMovePacketsQueue() {
		this.movePacketsQueue.clear();
	}

	// ----- HANDLE METHODS ----- //

	public void handlePacketKeepAlive(PacketInKeepAlive packet) {
		if (packet.getTime() == this.lastKeepAliveTick) {
			// Do something
		}
	}

	public void handlePacketLoginStart(PacketInLoginStart packet) {
		if (this.state != ConnectionState.NOTHING) {
			this.kick("You can't send this packet at this state!!! HACKER??");
			return;
		}

		this.state = ConnectionState.LOGIN;
		this.userName = packet.getUserName();

		if (packet.getProtocolVersion() != (byte)1) {
			this.kick("Please update your client!");
		}
	}

	public void handlePacketEncryptionKey(PacketInEncryptionKey packet) {
		if (this.state != ConnectionState.LOGIN) {
			this.kick("You can't send this packet at this state!!! HACKER???");
			return;
		}

		this.state = ConnectionState.AUTHENTICATION;
		try {
			byte[] encryptedSecretKey = packet.getEncryptedSecretKey();

			Cipher cipher = Cipher.getInstance("RSA/ECB/PKCS1Padding");
			cipher.init(Cipher.DECRYPT_MODE, PokemonServer.getServer().getServerKey().getPrivate());
			byte[] secretKeyData = cipher.doFinal(encryptedSecretKey);

			SecretKey secretKey = new SecretKeySpec(secretKeyData, "AES");
			this.netManager.enableEncryption(secretKey);
			this.sendPacket(new PacketOutAuthenticationRequest(this.netManager.getRemoteAddress().toString()));
		} catch (Exception ex) {
			PokemonServer.getServer().getLogger().warn("Bad public key", ex);
			this.kick("Bad public key");
		}
	}

	public void handlePacketAuthenticationResponse(final PacketInAuthenticationResponse packet) {
		if (this.state != ConnectionState.AUTHENTICATION) {
			this.kick("You can't send this packet at this state!!! HACKER???");
			return;
		}

		// Load player from database
		PokemonServer.getServer().getDatabaseManager().getPlayer(this.getUserName(), new DatabaseCallback<DatabasePlayer>() {
			@Override
			public void run(DatabasePlayer dbPlayer) {
				if (dbPlayer == null) {
					PokemonServer.getServer().getLogger().info("Wrong username");
					ClientHandle.this.kick("Der Benutzername existiert nicht");
					return;
				}

				String ipHashed = CryptManager.createMD5(dbPlayer.getPasswordHash() + ClientHandle.this.getNetworkManager().getRemoteAddress().toString());
				if (!ipHashed.equals(packet.getPasswordHash())) {
					PokemonServer.getServer().getLogger().info("Wrong password");
					ClientHandle.this.kick("Ungültiges Passwort");
					return;
				}

				if (dbPlayer.getLocation() == null) {
					dbPlayer.setLocation(new Location("Hauptwelt", 296, 102));
				}

				// Create player
				ClientHandle.this.player = new Player(dbPlayer.getLocation(), ClientHandle.this);
				PokemonServer.getServer().onPlayerJoin(ClientHandle.this.player);
			}
		});
	}

	public void handlePacketMove(InPacketMove packet) {
		if (this.state != ConnectionState.INGAME) {
			this.kick("You can't send this packet at this state!!! HACKER???");
			return;
		}

		this.movePacketsQueue.add(packet);

		if (this.movePacketsQueue.size() > 6) {
			// Too many move packets
			this.movePacketsQueue.clear();
			this.sendPacket(new PacketOutEntityTeleport(this.player.getEntityID(), this.player.getLocation(), this.player.getFace()));
		}
	}

	public void handleMove(PacketInPlayerMove packet) {
		if (!packet.getCurrentLocation().equalsCoords(this.player.getLocation())) {
			this.movePacketsQueue.clear();
			this.sendPacket(new PacketOutEntityTeleport(this.player.getEntityID(), this.player.getLocation(), this.player.getFace()));
			return;
		}

		this.player.doMove(packet.getEntityFace());
	}

	public void handleChangeFace(PacketInChangeFace packet) {
		this.player.setFace(packet.getNewFace());
	}

	public void handleChangeState(PacketInChangeState packet) {
		this.player.setState(packet.getNewState());
	}

	////////////////////////////////

	public void onLoginSuccess() {
		if (this.state == ConnectionState.INGAME) {
			throw new RuntimeException("Already ingame!");
		}

		this.sendPacket(new PacketOutLoginSuccess(this.player));
		this.state = ConnectionState.INGAME;
		PokemonServer.getServer().getEventManager().callEvent(new PlayerJoinEvent(this.player));

		for (Player p : this.player.getWorld().getPlayers()) {
			if (p != this.player) {
				this.sendPacket(new PacketOutPlayerSpawn(p));
			}
		}
	}

	public void setState(ConnectionState state) {
		this.state = state;
	}

	public void onDisconnect(String message) {
		if (this.player != null)
			PokemonServer.getServer().onPlayerLeave(this.player);
	}

	public void kick(final String message) {
		this.netManager.scheduleOutboundPacket(new PacketOutDisconnect(message), new GenericFutureListener() {
			@Override
			public void operationComplete(Future future) throws Exception {
				ClientHandle.this.netManager.closeChannel(message);
			}
		});
		this.netManager.disableAutoRead();
	}

	public void sendPacket(OutPacket packet) {
		try {
			this.netManager.scheduleOutboundPacket(packet, new GenericFutureListener[0]);
		} catch (Exception ex) {
			PokemonServer.getServer().getLogger().warn("Can't send packet", ex);
		}
	}

	///--- API METHODS ---///
	public void fadeOutScreen(float duration) {
		this.sendPacket(new PacketOutFadeScreen(PacketOutFadeScreen.FadeType.FADE_OUT, duration));
	}

	public void fadeInScreen(float duration) {
		this.sendPacket(new PacketOutFadeScreen(PacketOutFadeScreen.FadeType.FADE_IN, duration));
	}

}
