package de.howaner.Pokemon.network;

import com.google.common.collect.Queues;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import de.howaner.Pokemon.network.netty.NettyEncryptingDecoder;
import de.howaner.Pokemon.network.netty.NettyEncryptingEncoder;
import de.howaner.Pokemon.network.packets.in.InPacket;
import de.howaner.Pokemon.network.packets.out.OutPacket;
import de.howaner.Pokemon.network.util.CryptManager;
import io.netty.channel.Channel;
import io.netty.channel.ChannelFutureListener;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.SimpleChannelInboundHandler;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.util.concurrent.GenericFutureListener;
import java.net.SocketAddress;
import java.util.Queue;
import java.util.concurrent.TimeoutException;
import javax.crypto.SecretKey;

public class NetworkManager extends SimpleChannelInboundHandler {
	public static final NioEventLoopGroup eventLoops = new NioEventLoopGroup(0, (new ThreadFactoryBuilder()).setNameFormat("Netty IO #%d").setDaemon(true).build());

	private final Queue<InPacket> inPacketsQueue = Queues.newConcurrentLinkedQueue();
	private final Queue<QueuedOutPacket> outPacketsQueue = Queues.newConcurrentLinkedQueue();

	/**The active channel */
	private Channel channel;

	/** The address of the remote party */
	private SocketAddress socketAddress;

	/** The client handle */
	private ClientHandle clientHandle;

	/** A String indicating why the network has shutdown. */
	private String disconnectMessage = "";

	public NetworkManager() {
		this.clientHandle = new ClientHandle(this);
	}

	public ClientHandle getClientHandle() {
		return this.clientHandle;
	}

	public boolean isChannelOpen() {
		return (this.channel != null && this.channel.isOpen());
	}

	public String getDisconnectMessage() {
		return this.disconnectMessage;
	}

	public SocketAddress getRemoteAddress() {
		return this.socketAddress;
	}

	public void disableAutoRead() {
		this.channel.config().setAutoRead(false);
	}

	@Override
	public void channelActive(ChannelHandlerContext ctx) throws Exception {
		super.channelActive(ctx);
		this.channel = ctx.channel();
		this.socketAddress = ctx.channel().remoteAddress();
		this.channel.config().setAutoRead(true);
	}

	@Override
	public void channelInactive(ChannelHandlerContext ctx) {
		this.closeChannel("Ende der Datenübertragung oder so");
	}

	@Override
	public void exceptionCaught(ChannelHandlerContext ctx, Throwable cause) {
		String kickMessage;

		if (cause instanceof TimeoutException) {
			kickMessage = "Du bischd ausgetimd! Kauf besseres Internet!";
		} else {
			kickMessage = "Internal Exception: " + cause;
		}

		this.closeChannel(kickMessage);
	}

	/** Closes the channel, the parameter can be used for an exit message (not certain how it gets sent) */
	public void closeChannel(String kickMessage) {
		if (this.channel.isOpen()) {
			this.disconnectMessage = kickMessage;
			this.channel.close();
		}
	}

	@Override
	protected void channelRead0(ChannelHandlerContext ctx, Object msg) throws Exception {
		if (!this.channel.isOpen()) {
			return;
		}
		InPacket packet = (InPacket) msg;

		if (packet.hasPriority()) {
			packet.handlePacket(this.clientHandle);
		} else {
			this.inPacketsQueue.add(packet);
		}
	}

	/** Called every tick */
	public void processQueue() {
		this.flushOutboundQueue();
		int currentPacket = 0;
		while (!this.inPacketsQueue.isEmpty() && currentPacket < 1000) {
			InPacket packet = this.inPacketsQueue.poll();
			packet.handlePacket(this.clientHandle);

			currentPacket++;
		}

		this.clientHandle.onNetworkTick();
		this.channel.flush();
	}

	private void flushOutboundQueue() {
		if (this.channel != null && this.channel.isOpen()) {
			while (!this.outPacketsQueue.isEmpty()) {
				QueuedOutPacket queuedPacket = this.outPacketsQueue.poll();
				this.dispatchPacket(queuedPacket.packet, queuedPacket.listeners);
			}
		}
	}

	public void scheduleOutboundPacket(OutPacket packet, GenericFutureListener ... listeners) {
		if (this.channel != null && this.channel.isOpen()) {
			this.flushOutboundQueue();
			this.dispatchPacket(packet, listeners);
		} else {
			this.outPacketsQueue.add(new QueuedOutPacket(packet, listeners));
		}
	}

	private void dispatchPacket(final OutPacket packet, final GenericFutureListener[] listeners) {
		if (this.channel.eventLoop().inEventLoop()) {
			// Sync sending
			this.channel.writeAndFlush(packet).addListeners(listeners).addListener(ChannelFutureListener.FIRE_EXCEPTION_ON_FAILURE);
		} else {
			// Async sending
			this.channel.eventLoop().execute(new Runnable() {
				@Override
				public void run() {
					NetworkManager.this.channel.writeAndFlush(packet).addListeners(listeners).addListener(ChannelFutureListener.FIRE_EXCEPTION_ON_FAILURE);
				}
			});
		}
	}

	public void enableEncryption(SecretKey key) throws Exception {
		this.channel.pipeline().addBefore("splitter", "decrypt", new NettyEncryptingDecoder(CryptManager.createDecryptCipher(key)));
		this.channel.pipeline().addBefore("prepender", "encrypt", new NettyEncryptingEncoder(CryptManager.createEncryptCipher(key)));
	}

	public class QueuedOutPacket {
		public final OutPacket packet;
		public final GenericFutureListener[] listeners;

		public QueuedOutPacket(OutPacket packet, GenericFutureListener[] listeners) {
			this.packet = packet;
			this.listeners = listeners;
		}
	}

}
