package de.howaner.Pokemon.world;

import com.google.common.collect.Queues;
import de.howaner.Pokemon.PokemonServer;
import de.howaner.Pokemon.map.MapTerrain;
import de.howaner.Pokemon.saveformat.TagCompound;
import de.howaner.Pokemon.saveformat.TagManager;
import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Queue;

public class WorldLoader extends Thread {
	private Queue<WorldLoadTask> tasks = Queues.newConcurrentLinkedQueue();
	private Queue<WorldLoadedTask> loadedMaps = Queues.newConcurrentLinkedQueue();

	public WorldLoader() {
		super("WorldLoader");
	}

	public void onTick(WorldManager worldManager) {
		while (!this.loadedMaps.isEmpty()) {
			WorldLoadedTask task = this.loadedMaps.poll();
			if (task != null) {
				worldManager.addLoadedWorld(task.world);

				for (WorldLoadedCallback callback : task.callbacks) {
					callback.onLoadedWorld(task.world);
				}
			}
		}
	}

	@Override
	public void run() {
		PokemonServer server = PokemonServer.getServer();
		while (server.isRunning() && !this.isInterrupted()) {
			if (tasks.isEmpty()) {
				// Wait 1 second
				try {
					Thread.sleep(1000L);
				} catch (Exception e) {}

				continue;
			}

			WorldLoadTask task = tasks.poll();
			if (task != null) {
				this.loadWorld(task);
			}
		}
	}

	private void loadWorld(WorldLoadTask task) {
		try {
			TagCompound compound = TagManager.readFromFile(new File("maps/", task.worldName + ".pkmap"));

			MapTerrain terrain = new MapTerrain();
			terrain.setName(task.worldName);
			terrain.loadFromCompound(compound);

			World world = new World(terrain);
			this.loadedMaps.add(new WorldLoadedTask(world, task.callback));
		} catch (Exception ex) {
			PokemonServer.getServer().getLogger().error("Can't load world " + task.worldName, ex);
		}
	}

	public void addWorldToQueue(String worldName) {
		addWorldToQueue(worldName, null);
	}

	public void addWorldToQueue(String worldName, WorldLoadedCallback callback) {
		if (PokemonServer.getServer().getWorldManager().getWorld(worldName) != null) {
			if (callback != null)
				callback.onLoadedWorld(PokemonServer.getServer().getWorldManager().getWorld(worldName));
			return;
		}

		WorldLoadTask task = isWorldInQueue(worldName);
		if (task != null) {
			// World is already loading -> Add only the callback
			if (callback != null) {
				synchronized (task.callback) {
					task.callback.add(callback);
				}
			}
			return;
		}

		WorldLoadedTask loadedTask = isWorldInLoadedMapsQueue(worldName);
		if (loadedTask != null) {
			// World is already loaded -> Add only the callback
			if (callback != null) {
				synchronized (loadedTask.callbacks) {
					loadedTask.callbacks.add(callback);
				}
			}
			return;
		}

		// World isn't in the load queue -> Add world to the queue
		if (callback == null) {
			task = new WorldLoadTask(worldName);
		} else {
			task = new WorldLoadTask(worldName, callback);
		}

		this.tasks.add(task);
	}

	public WorldLoadTask isWorldInQueue(String worldName) {
		Iterator<WorldLoadTask> itr = this.tasks.iterator();
		while (itr.hasNext()) {
			WorldLoadTask task = itr.next();
			if (task.worldName.equals(worldName)) {
				return task;
			}
		}
		return null;
	}

	public WorldLoadedTask isWorldInLoadedMapsQueue(String worldName) {
		Iterator<WorldLoadedTask> itr = this.loadedMaps.iterator();
		while (itr.hasNext()) {
			WorldLoadedTask task = itr.next();
			if (task.world.getName().equals(worldName)) {
				return task;
			}
		}
		return null;
	}

	public static class WorldLoadTask {
		public final String worldName;
		public final List<WorldLoadedCallback> callback;

		public WorldLoadTask(String worldName, List<WorldLoadedCallback> callback) {
			this.worldName = worldName;
			this.callback = callback;
		}

		public WorldLoadTask(String worldName, WorldLoadedCallback callback) {
			this.worldName = worldName;
			this.callback = new ArrayList<WorldLoadedCallback>();
			this.callback.add(callback);
		}

		public WorldLoadTask(String worldName) {
			this.worldName = worldName;
			this.callback = new ArrayList<WorldLoadedCallback>();
		}
	}

	public static class WorldLoadedTask {
		public final World world;
		public final List<WorldLoadedCallback> callbacks;

		public WorldLoadedTask(World world, List<WorldLoadedCallback> callbacks) {
			this.world = world;
			this.callbacks = callbacks;
		}
	}

	public interface WorldLoadedCallback {
		public void onLoadedWorld(World world);
	}

}
